r"""

**openpnm.models.phases**

----

This submodule contains models for calculating thermophysical properties of
liquids, gases, and solids.

"""

from . import density
from . import diffusivity
from . import electrical_conductivity
from . import misc
from . import mixtures
from . import molar_density
from . import partition_coefficient
from . import permittivity
from . import surface_tension
from . import thermal_conductivity
from . import vapor_pressure
from . import viscosity
